/**************************************************************************************

EksLOKI89 - some variation of LOKI89 with 256 bit key, 64 bit block and 18 rounds

Implemented by Alexander Myasnikow

Web: www.darksoftware.narod.ru

**************************************************************************************/


#include <stdio.h>
#include <stdlib.h>

typedef unsigned long u32;
typedef unsigned char u8;

u32 f3 (u32, u32, u32, u32);
u32 f4 (u32, u32, u32, u32, u32);

#include <string.h>
#include "loki.i"		/* include LOKI89 Permutation, Subs & Key tables */

/*
 *      perm32(out, in, perm) is the general permutation of a 32-bit input
 *              block to a 32-bit output block, under the control of a
 *              permutation array perm. Each element of perm specifies which
 *              input bit is to be permuted to the output bit with the same
 *              index as the array element.
 *
 *      nb: to set bits in the output word, as mask with a single 1 in it is
 *              used. On each step, the 1 is shifted into the next location
 */

#define MSB     0x80000000L	/* MSB of 32-bit word */



u8
rol8 (u8 x)
{
  u8 result;
  __asm
  {
    mov al, x;
    mov cl, 1;
    rol al, cl;
    mov result, al;
  }

  return result;
}


void
perm32 (out, in, perm)
     u32 *out;			/* Output 32-bit block to be permuted                */
     u32 *in;			/* Input  32-bit block after permutation             */
     char perm[32];		/* Permutation array                                 */
{
  u32 mask = MSB;		/* mask used to set bit in output    */
  register int i, o, b;		/* input bit no, output bit no, value */
  register u8 *p = perm;	/* ptr to permutation array  */

  *out = 0;			/* clear output block */
  for (o = 0; o != 32; o++)	/* For each output bit position o */
    {
      i = (int) *p++;		/* get input bit permuted to output o */
      b = (*in >> i) & 01;	/* value of input bit i */
      if (b)			/* If the input bit i is set */
	*out |= mask;		/*  OR in mask to output i */
      mask >>= 1;		/* Shift mask to next bit    */
    }
}



static short s ();		/* declare LOKI89 s-box routine (for small ver) */

u32 lokikey[64];
u8 *bytekey = (u8 *) & lokikey[0];
u8 extkey[32][256];
u8 *ek = &extkey[0][0];



void __stdcall
setup (u8 * key)
{
  u32 i, n;

  u8 s, t;

  for (i = 0; i != 32; i++)
    {
      for (n = 0; n != 256; n++)
	{
	  extkey[i][n] = EBOX[n ^ key[i]];
	}
    }

  for (i = 0; i != 256; i++)
    {
      bytekey[i] = i;
    }

  s = 0;

  for (n = 0; i != 49152; i++)
    {
      s = bytekey[rol8 (EBOX[s] + bytekey[i & 0xff] + ek[i % 8192])];
      t = bytekey[i & 0xff];
      bytekey[i & 0xff] = bytekey[s & 0xff];
      bytekey[s & 0xff] = t;

    }

}



void __stdcall
crypt (u32 * b)
{
  register u32 L, R;

  L = b[0] ^ lokikey[0];
  R = b[1] ^ lokikey[1];

  L ^= f4 (R, lokikey[2], lokikey[20], lokikey[21], lokikey[56]);
  R ^= f3 (L, lokikey[3], lokikey[22], lokikey[23]);
  L ^= f3 (R, lokikey[4], lokikey[24], lokikey[25]);
  R ^= f4 (L, lokikey[5], lokikey[26], lokikey[27], lokikey[57]);
  L ^= f3 (R, lokikey[6], lokikey[28], lokikey[29]);
  R ^= f3 (L, lokikey[7], lokikey[30], lokikey[31]);
  L ^= f4 (R, lokikey[8], lokikey[32], lokikey[33], lokikey[58]);
  R ^= f3 (L, lokikey[9], lokikey[34], lokikey[35]);
  L ^= f3 (R, lokikey[10], lokikey[36], lokikey[37]);
  R ^= f4 (L, lokikey[11], lokikey[38], lokikey[39], lokikey[59]);
  L ^= f3 (R, lokikey[12], lokikey[40], lokikey[41]);
  R ^= f3 (L, lokikey[13], lokikey[42], lokikey[43]);
  L ^= f4 (R, lokikey[14], lokikey[44], lokikey[45], lokikey[60]);
  R ^= f3 (L, lokikey[15], lokikey[46], lokikey[47]);
  L ^= f3 (R, lokikey[16], lokikey[48], lokikey[49]);
  R ^= f4 (L, lokikey[17], lokikey[50], lokikey[51], lokikey[61]);
  L ^= f3 (R, lokikey[18], lokikey[52], lokikey[53]);
  R ^= f3 (L, lokikey[19], lokikey[54], lokikey[55]);

  b[0] = R ^ lokikey[62];
  b[1] = L ^ lokikey[63];

}


void __stdcall
decrypt (u32 * b)
{

  register u32 L, R;

  R = b[0] ^ lokikey[62];
  L = b[1] ^ lokikey[63];

  R ^= f3 (L, lokikey[19], lokikey[54], lokikey[55]);
  L ^= f3 (R, lokikey[18], lokikey[52], lokikey[53]);
  R ^= f4 (L, lokikey[17], lokikey[50], lokikey[51], lokikey[61]);
  L ^= f3 (R, lokikey[16], lokikey[48], lokikey[49]);
  R ^= f3 (L, lokikey[15], lokikey[46], lokikey[47]);
  L ^= f4 (R, lokikey[14], lokikey[44], lokikey[45], lokikey[60]);
  R ^= f3 (L, lokikey[13], lokikey[42], lokikey[43]);
  L ^= f3 (R, lokikey[12], lokikey[40], lokikey[41]);
  R ^= f4 (L, lokikey[11], lokikey[38], lokikey[39], lokikey[59]);
  L ^= f3 (R, lokikey[10], lokikey[36], lokikey[37]);
  R ^= f3 (L, lokikey[9], lokikey[34], lokikey[35]);
  L ^= f4 (R, lokikey[8], lokikey[32], lokikey[33], lokikey[58]);
  R ^= f3 (L, lokikey[7], lokikey[30], lokikey[31]);
  L ^= f3 (R, lokikey[6], lokikey[28], lokikey[29]);
  R ^= f4 (L, lokikey[5], lokikey[26], lokikey[27], lokikey[57]);
  L ^= f3 (R, lokikey[4], lokikey[24], lokikey[25]);
  R ^= f3 (L, lokikey[3], lokikey[22], lokikey[23]);
  L ^= f4 (R, lokikey[2], lokikey[20], lokikey[21], lokikey[56]);

  b[0] = L ^ lokikey[0];
  b[1] = R ^ lokikey[1];

}



/*
 *      f(r, k) - is the complex non-linear LOKI89 function, whose output
 *              is a complex function of both input data and sub-key.
 *              The input data R(i-1) is:
 *                      added modulo 2 to the key K(i)
 *                      expanded to 48-bits via expansion fn E
 *                      substituted into the S-boxes
 *                      permuted by P.
 *              ie the calculation is:
 *                      A = E(R(i-1) XOR K(i))          (a 48-bit value)
 *                      B = S(A)                        (a 32-bit value)
 *                      f = P(B)                        (a 32-bit value)
 *
 *      nb: the 12-bit S-box input value [x x x x 11 10 9 ... 3 2 1 0]
 *          is interpreted as:
 *              bits [11 10 1 0] select 1 of 16 rows within each box,
 *              bits [9 8 ... 3 2] then select a column within that row
 */

#define MASK12  0x0fff		/* 12 bit mask for expansion E */

u32
f3 (u32 r, u32 k1, u32 k2, u32 k3)
{
  u32 a, b, c;			/* 32 bit S-box output, & P output */

  a = r ^ k1;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);

  perm32 (&c, &b, P);		/* C = P(S( E(R(i-1)) XOR K(i))) */

  a = c ^ k2;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);


  a = b ^ k3;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);



  perm32 (&c, &b, P);		/* C = P(S( E(R(i-1)) XOR K(i))) */

  return c;			/* f returns the result C */
}


u32
f4 (u32 r, u32 k1, u32 k2, u32 k3, u32 k4)
{
  u32 a, b, c;			/* 32 bit S-box output, & P output */

  a = r ^ k1;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);

  perm32 (&c, &b, P);		/* C = P(S( E(R(i-1)) XOR K(i))) */

  a = c ^ k2;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);


  a = b ^ k3;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);


  a = b ^ k4;			/* A = R(i-1) XOR K(i) */

  /* want to use slow speed/small size version */
  b = ((u32) s ((a & MASK12))) |	/* B = S(E(R(i-1))^K(i)) */
    ((u32) s (((a >> 8) & MASK12)) << 8) |
    ((u32) s (((a >> 16) & MASK12)) << 16) |
    ((u32) s ((((a >> 24) | (a << 8)) & MASK12)) << 24);



  perm32 (&c, &b, P);		/* C = P(S( E(R(i-1)) XOR K(i))) */

  return c;			/* f returns the result C */
}


/*
 *      s(i) - return S-box value for input i
 */
static short
s (i)
     register u32 i;		/* return S-box value for input i */
{
  register short r, c, v, t;
  short exp8 ();		/* exponentiation routine for GF(2^8) */

  r = ((i >> 8) & 0xc) | (i & 0x3);	/* row value */
  c = (i >> 2) & 0xff;		/* column value */
  t = c ^ r;			/* base value for Sfn */
  v = exp8 (t, sfn[r].exp, sfn[r].gen);	/* Sfn[r] = t ^ exp mod gen */
  return (v);
}


/*
 *      mult8(a, b, gen) - returns the product of two binary
 *              strings a and b using the generator gen as the modulus
 *                      mult = a * b mod gen
 *              gen generates a suitable Galois field in GF(2^8)
 */

#define SIZE 256		/* 256 elements in GF(2^8) */

short
mult8 (a, b, gen)
     short a, b;		/* operands for multiply */
     short gen;			/* irreducible polynomial generating Galois Field */
{
  short product = 0;		/* result of multiplication */

  while (b != 0)		/* while multiplier is non-zero */
    {
      if (b & 01)
	product ^= a;		/*   add multiplicand if LSB of b set */
      a <<= 1;			/*   shift multiplicand one place */
      if (a >= SIZE)
	a ^= gen;		/*   and modulo reduce if needed */
      b >>= 1;			/*   shift multiplier one place  */
    }
  return (product);
}

/*
 *      exp8(base, exponent, gen) - returns the result of
 *              exponentiation given the base, exponent, generator of GF,
 *                      exp = base ^ exp mod gen
 */

short
exp8 (base, exponent, gen)
     short base;		/* base of exponentiation       */
     short exponent;		/* exponent                     */
     short gen;			/* irreducible polynomial generating Galois Field */
{
  short accum = base;		/* superincreasing sequence of base */
  short result = 1;		/* result of exponentiation         */

  if (base == 0)		/* if zero base specified then      */
    return (0);			/* the result is "0" if base = 0    */

  while (exponent != 0)		/* repeat while exponent non-zero */
    {
      if ((exponent & 0x0001) == 0x0001)	/* multiply if exp 1 */
	result = mult8 (result, accum, gen);
      exponent >>= 1;		/* shift exponent to next digit */
      accum = mult8 (accum, accum, gen);	/* & square  */
    }
  return (result);
}




u32 __stdcall
getblocksize ()
{
  return 64;
}

u32 __stdcall
getkeysize ()
{
  return 256;
}

void __stdcall
getciphername (u8 * p)
{
  strcpy (p, "EksLOKI-89-256");
}
